from OpenGL import GL
import numpy as np

from .__prelude__ import (
    dynamic, SELF, Matrix, unit_square_to_image,
    Shader, PureShader, Qt, GLProgram, destroy_texture,
    Point
)
from .Scene import Scene, SceneDrawing

GEO_SHADER = """
#version 330
uniform mat3 geo_to_lut;
uniform sampler2D lut_texture;

vec4 {radar_color}(vec3 geo_coords, inout float up_opacity);

vec4 geo_color(vec3 geo_coords) {{
  vec3 lut_coords_homo = geo_to_lut * geo_coords;
  vec2 lut_coords = lut_coords_homo.xy / lut_coords_homo.z;
  vec4 lut_values = texture(lut_texture, lut_coords);
  float up_opacity = 1.0;
  return {radar_color}(vec3(lut_values.xy,1), up_opacity);
}}
"""

class GeoScene(Scene):
    selected_band    = dynamic.readonly()
    lut_geo_dataset  = dynamic.readonly()
    overlay_layer    = dynamic.external()
    tile_cache       = dynamic.external()

    def __init__(self, selected_band, lut_geo_dataset, dynamic_overlay_layer, dynamic_tile_cache):
        super().__init__()
        self._selected_band = selected_band
        self._lut_geo_dataset = lut_geo_dataset
        self._dynamic_tile_cache = dynamic_tile_cache
        self._dynamic_overlay_layer = dynamic_overlay_layer
        self.distance = 4.0

        self.dynamic_attribute("_rerender").value_changed.connect(lambda: self._rerender())

    @dynamic.method(SELF.model_to_world, SELF.tile_cache.provider)
    def _rerender(self):
        self.renderChanged.emit()

    @dynamic.memo(SELF.lut_geo_dataset.size, SELF.lut_geo_dataset.pixel_to_crs, SELF.model_to_image)
    def model_to_geo(self):
        w, h = self.lut_geo_dataset.size
        return Matrix.product(
            self.lut_geo_dataset.pixel_to_crs,
            Matrix.scale((w, h, 1)),
            self.model_to_image,
        )
    @dynamic.memo(SELF.distance, SELF.center)
    def model_to_world(self):
        return Matrix.product(
            Matrix.translate((0,0,self.distance)),
            Matrix.translate((self.center.x, self.center.y, 0.0)),
        )
    @dynamic.memo(SELF.lut_geo_dataset.size)
    def model_to_image(self):
        w, h = self.lut_geo_dataset.size
        return unit_square_to_image(w,h)

    @dynamic.memo(SELF.lut_geo_dataset)
    def lut_image(self):
        ret = np.stack([
            self.lut_geo_dataset.read(1),
            self.lut_geo_dataset.read(2),
        ], axis=-1).astype(np.float32)
        return ret
    @dynamic.memo(SELF.selected_band.size)
    def lut_scale(self):
        w, h = self.selected_band.size
        return np.array([1/w, 1/h])

    def whileDrag(self, dragMode, dragStart, dragEnd):
        if dragMode == Qt.Qt.KeyboardModifier.NoModifier:
            xa, ya = dragStart
            xb, yb = dragEnd
            self.center += Point(xb-xa, yb-ya)
        return dragStart

    @property
    def shaders(self):
        return [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex, "geo_map/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, "geo_map/fragment.glsl"),
        ]

    def initializeDrawing(self, context):
        return GeoSceneDrawing(context, self)

class GeoSceneDrawing(SceneDrawing):
    def __init__(self, context, geo_scene):
        super().__init__(context, geo_scene)
        self.__initialized_overlay = self.drawing.overlay_layer.GL_initialize(context)
        self.__initialized_overlay.dynamic_attribute("GL_setup").value_changed.connect(lambda: self.drawing.renderChanged.emit())

    @dynamic.memo(SELF.viewport_size)
    def world_to_clip(self):
        w, h = self.viewport_size
        hfactor = max(1.0, w/h)/4.0
        vfactor = max(1.0, h/w)/4.0
        near = 4.0*0.005
        return Matrix.frustum(-hfactor*near, hfactor*near,
                              -vfactor*near, vfactor*near,
                              -near, -1000.0)

    @dynamic.memo(SELF.drawing.overlay_layer.shaders)
    def GL_program(self):
        return GLProgram(
            self.drawing.shaders + [
                PureShader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, shader)
                for shader in self.drawing.overlay_layer.shaders
            ] + [
                PureShader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, GEO_SHADER.format(
                    radar_color = self.drawing.overlay_layer.local_ident("geo_color")
                ))
            ]
        )
    @dynamic.memo(SELF.GL_program)
    def GL_mesh(self):
        return self.GL_program.create_square_mesh(GL.GL_TRIANGLES)

    @dynamic.memo(SELF.drawing.lut_geo_dataset, destroy = destroy_texture)
    def GL_lut_geo_texture(self):
        ret = Qt.QOpenGLTexture(Qt.QOpenGLTexture.Target2D)
        w, h = self.drawing.lut_geo_dataset.size
        img = self.drawing.lut_image

        ret.setWrapMode(Qt.QOpenGLTexture.ClampToEdge)
        ret.setSize(w, h)
        ret.setFormat(Qt.QOpenGLTexture.TextureFormat.RG32F)
        ret.allocateStorage(Qt.QOpenGLTexture.PixelFormat.RG, Qt.QOpenGLTexture.PixelType.Float32)
        ret.setData(Qt.QOpenGLTexture.PixelFormat.RG, Qt.QOpenGLTexture.PixelType.Float32, img) #type: ignore
        ret.create()

        return ret

    @dynamic.memo(SELF.viewport_size)
    def tiles_per_screen(self):
        w,h = self.viewport_size
        return max(w/256.0, h/256.0) # Each tile is still 256 pixels wide

    def paintGL(self):
        gl = self.context.functions()

        program = self.GL_program
        mesh = self.GL_mesh

        lut_texture = self.GL_lut_geo_texture
        setup_overlay = self.__initialized_overlay.GL_setup

        gl.glEnable(GL.GL_BLEND)
        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        gl.glEnable(GL.GL_DEPTH_TEST)
        gl.glClearColor(0.0,0.0,0.0,0.0)
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        self.drawing.tile_cache.GL_paint_background(
            gl,
            model_to_crs = self.drawing.model_to_geo,
            clip_to_model = self.clip_to_model,
            model_to_world = self.drawing.model_to_world,
            world_to_clip = self.world_to_clip,
            tiles_per_screen = self.tiles_per_screen,
        )

        with program as p:
            setup_overlay(p)
            p.uniforms.lut_texture = lut_texture

            p.uniforms.model_to_clip = Matrix.product(
                self.world_to_clip,
                self.drawing.model_to_world,
            )
            p.uniforms.model_to_geo = self.drawing.model_to_geo
            p.uniforms.geo_to_lut = self.drawing.model_to_image * self.drawing.model_to_geo.inverse()
            mesh.draw()

    def freeGL(self):
        self.dynamic_attribute("GL_lut_geo_texture").clear()



import xyzservices.providers

from .__prelude__ import Qt
from .Schema import Schema

PROVIDERS = {
    "OpenStreetMap": xyzservices.providers["OpenStreetMap"]["Mapnik"],
    "USGS": xyzservices.providers["USGS"]["USImagery"],
    "Esri World Topo": xyzservices.providers["Esri"]["WorldTopoMap"],
}
for p in PROVIDERS:
    PROVIDERS[p].name = p

class TileProviderSchema(Schema):
    class TileProviderNode(Schema.Node):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema)
            self.set_model(model)
            self.__editor = Qt.QComboBox()
            for p in PROVIDERS:
                self.__editor.addItem(p, p)
            def on_change(index):
                data = self.__editor.itemData(index)
                self.requestModelChange.emit(PROVIDERS[data])
            self.__editor.currentIndexChanged.connect(on_change)

        def editor(self):
            return self.__editor
        def createEditor(self, parent):
            self.__editor.setParent(parent)
            return self.__editor
        def destroyEditor(self):
            return True
        def setEditorData(self, editor):
            # TODO
            return True
        def setModelData(self, editor):
            data = editor.itemData(editor.currentIndex())
            self.requestModelChange.emit(PROVIDERS[data])
            return True

        def set_model(self, model):
            super().set_model(model)
            self.dataChanged.emit([])

    def make_node(self, item_model, model):
        return self.TileProviderNode(item_model,self,model)

    def supported_mime_types(self):
        return set(["application/x-insarviz-tile_provider"])
    def can_model_from_mime(self, mimeData):
        return mimeData.hasFormat("application/x-insarviz-tile_provider")
    def model_from_mime(self, mimeData):
        data_bytes = mimeData.data("application/x-insarviz-tile_provider").data()
        provider = json.loads(bytearray(data_bytes))
        return PROVIDERS[provider['name']]
    def can_mime_from_model(self, model):
        return True
    def mime_from_model(self, model):
        mimeData = QMimeData()
        dic = {
            "name": model.name
        }
        mimeData.setData("application/x-insarviz-tile_provider", bytes(json.dumps(dic), "utf-8"))
        return mimeData



import qasync
import asyncio
import pathlib
import sys
import rasterio as rio
import pyqtgraph
import argparse
import logging

from .__prelude__ import logger
from .state import WindowState, Dataset
from .view import InsarvizWindow
from .misc import Qt, GLOBAL_THREAD_POOL

def get_args():
    parser = argparse.ArgumentParser(
        prog = "ts_viz",
        description = "InSAR timeseries visualisation"
    )
    parser.add_argument("-l", type=str, default='INFO',
                        help=("set logging level (one of DEBUG, INFO, WARNING, ERROR, CRITICAL) [default INFO]"))
    parser.add_argument("-i", type=str, default=None, help="input filepath")
    return parser.parse_args()
args = get_args()
logger.setLevel(logging.__dict__.get(args.l, logging.INFO))

icon_path = pathlib.Path(__file__).parent / "icons"
logger.info("Setting icon search path to %s", icon_path)
Qt.QDir.addSearchPath('insarviz', icon_path)

class InsarvizApplication:
    def __init__(self):
        default_opengl_format = Qt.QSurfaceFormat.defaultFormat()
         # Necessary for tesselation shaders on Windows where the default OpenGL version is 3.x
        default_opengl_format.setVersion(4, 1)
        default_opengl_format.setProfile(Qt.QSurfaceFormat.OpenGLContextProfile.CoreProfile)
        default_opengl_format.setSwapBehavior(Qt.QSurfaceFormat.SwapBehavior.DoubleBuffer)
        default_opengl_format.setOption(Qt.QSurfaceFormat.FormatOption.DeprecatedFunctions, False)
        Qt.QSurfaceFormat.setDefaultFormat(default_opengl_format)

        Qt.QCoreApplication.setAttribute(Qt.Qt.ApplicationAttribute.AA_ShareOpenGLContexts)
        self.app = Qt.QApplication()
        self.app.setApplicationName("InsarViz")
        self.event_loop = qasync.QEventLoop(self.app)

    def run(self):
        with self.event_loop:
            self.event_loop.run_forever()

logger.info("Initializing Qt Application")
app = InsarvizApplication()
pyqtgraph.setConfigOption("foreground", 'k')
pyqtgraph.setConfigOption("background", 'w')
logger.info("Creating InsarViz window")
win_state = WindowState()
win = InsarvizWindow(win_state)

if args.i is not None:
    # Load the dataset after a small delay, just because we can
    timer = Qt.QTimer()
    timer.setInterval(100)
    timer.setSingleShot(True)
    def load_dataset():
        path = pathlib.Path(args.i)
        logger.debug('Opening file %s', path)
        win.open_file(path)
        logger.debug('Finished opening %s', path)
    timer.timeout.connect(load_dataset)
    timer.start()

win.show()
app.run()
win_state.exit()
logger.info("Exiting InSARViz. Please come back soon !")

GLOBAL_THREAD_POOL.abort_all()



import numpy as np
from OpenGL import GL

from .__prelude__ import (
    logger,
    Matrix, Qt, Bound, inOpenGLContext,
    dynamic, SELF, EACH,
    OpenGLPainter, destroy_texture
)
from .Layer import Layer, InitializedLayer, ExistingTexture

POINTS_SHADER = """
#version 330

uniform sampler2D {points_texture};
uniform float {opacity};
uniform mat3 {geo_to_image};

float {geo_height}(vec3 geo_coords) {{
   return 0.0;
}}

vec4 {geo_color}(vec3 geo_coords, inout float up_opacity) {{
  vec3 image_coords = {geo_to_image} * geo_coords;
  vec4 color = texture({points_texture}, image_coords.xy / image_coords.z);
  return vec4(color.xyz, color.w*{opacity}*up_opacity);
}}
"""

class PointsLayer(Layer):
    is_removable = False

    opacity     = dynamic.variable(1.0)
    is_enabled  = dynamic.variable(True)
    dataset     = dynamic.external()
    points      = dynamic.readonly()
    profiles    = dynamic.readonly()
    shaders     = dynamic.readonly()

    point_drawings    = dynamic.external()
    profile_drawings  = dynamic.external()

    def __init__(self, dyn_dataset, points, profiles):
        super().__init__()
        self._dynamic_dataset = dyn_dataset
        self._points: ObservableList[MapPoint] = points
        self._dynamic_point_drawings = (SELF.points / EACH(SELF.draw))[self]
        self._profiles: ObservableList[MapPoint] = profiles
        self._dynamic_profile_drawings = (SELF.profiles / EACH(SELF.draw))[self]

        self._shaders = [
            POINTS_SHADER.format(
                **self.local_idents("geo_color", "geo_height", "points_texture", "geo_to_image", "opacity")
            )
        ]

        self.dynamic_attribute("_redraw").value_changed.connect(lambda: self._redraw())

    @dynamic.method(SELF.dataset, SELF.GL_draw_points_and_profiles)
    def _redraw(self):
        self.renderChanged.emit()

    @dynamic.memo(SELF.dataset)
    def texture_size(self):
        return self.dataset.size
    @dynamic.memo(SELF.dataset, SELF.texture_size)
    def geo_to_image(self):
        w, h = self.texture_size
        return Matrix.product(
            self.dataset.pixel_to_crs,
            Matrix.scale((w,h,1))
        ).inverse()

    @dynamic.method(SELF.opacity, SELF.is_enabled, SELF.point_drawings, SELF.profile_drawings)
    def GL_draw_points_and_profiles(self, painter):
        painter.setPen(Qt.QColor(255, 255, 255))
        for p in self.profiles:
            p.draw(painter, self.opacity)
        for p in self.points:
            p.draw(painter, self.opacity)

    def GL_initialize(self, context):
        return InitializedPointsLayer(context, self)

    __mime_type__ = "x-application/insarviz/PointsLayer"
    @classmethod
    def can_from_mime(cls, mimeData, **kwargs):
        return super().can_from_mime(mimeData, **kwargs) and 'points' in kwargs and 'profiles' in kwargs and 'selected_band' in kwargs
    @staticmethod
    def from_dict(dct, points = None, profiles = None, selected_band = None, **__kwargs__):
        ret = PointsLayer(selected_band.dynamic_attribute("dataset"), points, profiles)
        ret.opacity = dct['opacity']
        ret.is_enabled = dct.get('is_enabled', True)
        return ret

    def to_dict(self):
        return {
            "opacity": self.opacity,
            "is_enabled": self.is_enabled,
        }

class InitializedPointsLayer(InitializedLayer):
    def __init__(self, context, points_layer):
        super().__init__(context, points_layer)
        self.__points_context = self.GL_create_shared_context()

    @dynamic.memo(SELF.layer.GL_draw_points_and_profiles,
                     destroy = destroy_texture)
    def GL_texture(self):
        return self.GL_paint_into_texture(*self.layer.texture_size, self.__points_context, self.layer.GL_draw_points_and_profiles)

    @dynamic.memo(SELF.GL_texture, SELF.layer.geo_to_image, SELF.layer.opacity, SELF.layer.is_enabled)
    def GL_setup(self):
        tex = self.GL_texture
        def doit(program):
            uni = self.local_uniforms(program.uniforms)
            uni.points_texture = tex
            uni.geo_to_image = self.layer.geo_to_image
            uni.opacity = self.layer.opacity if self.layer.is_enabled else 0.0
        return doit

    def GL_free(self):
        self.dynamic_attribute("GL_texture").clear()
        self.dynamic_attribute("GL_setup").clear()



import logging
import pathlib
import os

logger = logging.getLogger('insarviz')

INSARVIZ_ROOT = str(pathlib.Path(__file__).parent.parent) + os.sep

class InsarvizFilter:
    def filter(self, record):
        return record.name.startswith("insarviz")
logger.addFilter(InsarvizFilter())
logger.setLevel(logging.DEBUG)

class InsarvizFormatter(logging.Formatter):
    def __init__(self):
        super().__init__('%(asctime)s.%(msecs)03d %(levelname)s %(modulepath)s:%(lineno)d: %(message)s',
                         datefmt = '%Y-%m-%d %H:%M:%S')
    def format(self, record):
        record.modulepath = record.pathname.removeprefix(INSARVIZ_ROOT).removesuffix(".py").replace(os.sep, ".")
        return super().format(record)

log_handler = logging.StreamHandler()
log_handler.setFormatter(InsarvizFormatter())
logger.addHandler(log_handler)



from typing import override
import numpy as np
from . import Qt, Matrix
from OpenGL import GL

def unit_square_to_image(w, h):
    hfactor = max(float(h)/float(w), 1.0)
    vfactor = max(float(w)/float(h), 1.0)
    return Matrix.product(
        Matrix.translate((0.5,0.5)),
        Matrix.scale((hfactor, vfactor, 1.0)),
        Matrix.scale((0.5,-0.5,1.0))
    )

class DEMTexture(Qt.QOpenGLTexture):
    def __init__(self, image):
        super().__init__(Qt.QOpenGLTexture.Target.Target2D)
        self.image = image
        h, w, _ = self.image.shape
        self.texture_to_image = Matrix.product(
            unit_square_to_image(w, h),
            Matrix.translate((-1,-1)),
            Matrix.scale((2,2,1))
        )

    def create(self):
        if self.image is not None:
            self.setMagnificationFilter(Qt.QOpenGLTexture.Filter.Linear)
            self.setMinificationFilter(Qt.QOpenGLTexture.Filter.LinearMipMapLinear)
            self.setWrapMode(Qt.QOpenGLTexture.WrapMode.ClampToBorder)
            self.setBorderColor(Qt.QColor(0.0,0.0,0.0,0.0))

            h, w, _ = self.image.shape
            self.setSize(w, h)

            self.setFormat(Qt.QOpenGLTexture.TextureFormat.RGBA32F)
            self.allocateStorage(Qt.QOpenGLTexture.PixelFormat.RGBA, Qt.QOpenGLTexture.PixelType.Float32)
            self.setData(Qt.QOpenGLTexture.PixelFormat.RGBA, Qt.QOpenGLTexture.PixelType.Float32, self.image) #type: ignore
            self.generateMipMaps()

            self.image = None
        return super().create()



from OpenGL import GL
import numpy as np

from .__prelude__ import (Qt, dynamic, SELF, Matrix, destroy_texture)
from .Layer import Layer, InitializedLayer

SELECTED_BAND_SHADER = """
#version 330

uniform sampler2D {texture};
uniform sampler2D {reference_texture};
uniform sampler1D {colormap};
uniform float {opacity};
uniform mat3 {geo_to_texture};
uniform mat2 {value_to_color};
uniform float {dem_scale};

float {geo_height}(vec3 geo_coords) {{
   vec3 image_coords = {geo_to_texture} * geo_coords;
   vec4 value = texture({texture}, image_coords.xy / image_coords.z);
   vec4 ref_value = texture({reference_texture}, image_coords.xy / image_coords.z);
   value -= ref_value * ref_value.w;
   return clamp(-value.x * {dem_scale}, -1, 1);
}}

vec4 {geo_color}(vec3 geo_coords, inout float up_opacity) {{
  vec3 image_coords = {geo_to_texture} * geo_coords;
  vec4 value = texture({texture}, image_coords.xy / image_coords.z);
  vec4 ref_value = texture({reference_texture}, image_coords.xy / image_coords.z);
  value -= ref_value * ref_value.w;
  vec2 colormap_index = {value_to_color} * vec2(value.x, 1.0);
  vec4 color = texture({colormap}, colormap_index.x / colormap_index.y);
  return vec4(color.xyz, value.w*{opacity}*up_opacity);
}}
"""

class SelectedBandLayer(Layer):
    is_removable = False

    colormap       = dynamic.readonly()
    selected_band  = dynamic.readonly()
    shaders        = dynamic.readonly()
    opacity        = dynamic.variable(1.0)
    is_enabled     = dynamic.variable(True)
    dem_weight     = dynamic.variable(1.0)

    def __init__(self, selected_band, colormap):
        super().__init__()
        self._selected_band: SelectedBand = selected_band
        self._colormap: ColorMap = colormap
        self._shaders = [
            SELECTED_BAND_SHADER.format(
                **self.local_idents("geo_color", "geo_height", "texture", "reference_texture", "colormap", "opacity", "geo_to_texture", "value_to_color", "dem_scale")
            )
        ]

        self.dynamic_attribute("_redraw").value_changed.connect(lambda: self._redraw())

    @dynamic.memo(SELF.selected_band.image_to_crs)
    def geo_to_image(self):
        return self.selected_band.image_to_crs.inverse()

    @dynamic.method(SELF.colormap.createTexture, SELF.colormap.value_to_color,
                    SELF.selected_band.texture,
                    SELF.selected_band.reference_texture,
                    SELF.opacity, SELF.is_enabled, SELF.dem_weight, SELF.value_scale)
    def _redraw(self):
        self.renderChanged.emit()

    @dynamic.memo(SELF.selected_band.image)
    def value_scale(self):
        img = self.selected_band.image[:,:,0]
        a, b = np.nanquantile(img, 0.02), np.nanquantile(img, 0.98)
        sc = max(abs(a), abs(b))
        if sc == 0.0:
            return 1
        else:
            return 1/sc

    def GL_initialize(self, context: Qt.QOpenGLContext, /):
        return InitializedSelectedBandLayer(context, self)

    __mime_type__ = "x-application/insarviz/SelectedBandLayer"
    @classmethod
    def can_from_mime(cls, mimeData, **kwargs):
        return super().can_from_mime(mimeData, **kwargs) and 'selected_band' in kwargs
    @staticmethod
    def from_dict(dct, selected_band = None, colormap = None, **__kwargs__):
        ret = SelectedBandLayer(selected_band, colormap)
        ret.opacity = dct['opacity']
        ret.is_enabled = dct.get('is_enabled', True)
        ret.dem_weight = dct.get('dem_weight', 0.0)
        return ret
    def to_dict(self):
        return {
            "opacity": self.opacity,
            "is_enabled": self.is_enabled,
            "dem_weight": self.dem_weight,
        }

class InitializedSelectedBandLayer(InitializedLayer):
    def __init__(self, context, band_layer):
        super().__init__(context, band_layer)

    @dynamic.memo(SELF.layer.colormap.createTexture,
                     destroy = destroy_texture)
    def GL_colormap_texture(self):
        return self.layer.colormap.createTexture()

    @dynamic.memo(
        SELF.layer.selected_band.texture, SELF.layer.selected_band.reference_texture,
        SELF.GL_colormap_texture,
        SELF.layer.opacity, SELF.layer.colormap.value_to_color,
        SELF.layer.geo_to_image, SELF.layer.is_enabled, SELF.layer.value_scale
    )
    def GL_setup(self):
        tex = self.layer.selected_band.texture
        ref_tex = self.layer.selected_band.reference_texture
        colormap = self.GL_colormap_texture
        def doit(program):
            uni = self.local_uniforms(program.uniforms)
            uni.texture = tex
            uni.reference_texture = ref_tex
            uni.colormap = colormap
            uni.opacity = self.layer.opacity if self.layer.is_enabled else 0.0
            uni.geo_to_texture = self.layer.geo_to_image
            uni.value_to_color = self.layer.colormap.value_to_color
            uni.dem_scale = float(self.layer.dem_weight * self.layer.value_scale)
        return doit

    def GL_free(self):
        self.dynamic_attribute("GL_colormap_texture").clear()
        self.dynamic_attribute("GL_setup").clear()



import os

from .__prelude__ import (
    logger, Qt,
    WindowState, MaybeBandNumber, MapPoint, MapProfile,
    SwipeLayer, PointsLayer, SelectedBandLayer, RasterLayer, RasterRGBLayer, Layer,
    Dataset,
)
from .schema import (
    SchemaView, StructSchema, MaybeStructSchema, StructRoles, IntSchema, IntSliderSchema,
    ColorSchema, ListSchema, ListRoles, ListIconRoles, UnionSchema,
    FloatSliderSchema, FloatSchema, StringSchema, ColorMapSchema,
    TileProviderSchema
)
from .GLWidget import GLWidget
from .RasterLayerDialog import RasterLayerDialog
from .WidgetTree import Container, Leaf

class MapPointRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled | Qt.Qt.ItemFlag.ItemIsEditable | Qt.Qt.ItemFlag.ItemIsUserCheckable
    @staticmethod
    def field_roles(field):
        if field == "color":
            return [Qt.Qt.ItemDataRole.DecorationRole]
        if field == "name":
            return [Qt.Qt.ItemDataRole.DisplayRole]
        if field == "show_in_map":
            return [Qt.Qt.ItemDataRole.CheckStateRole]
        return None

    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return model.name
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return model.color
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            return Qt.Qt.CheckState.Checked if model.show_in_map else Qt.Qt.CheckState.Unchecked
    @staticmethod
    def set_role(model, role, value):
        if role == Qt.Qt.ItemDataRole.EditRole:
            model.name = value
            return True
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            model.show_in_map = (value == Qt.Qt.CheckState.Checked.value)
            return True
        return False

    @staticmethod
    def context_menu(parent_widget, node):
        ret = Qt.QMenu()
        node._action1 = Qt.QAction("Remove")
        def remove_self():
            me = node.self_index()
            if me.isValid():
                me.model().removeRow(me.row(), node.parent_index())
        node._action1.triggered.connect(remove_self)
        ret.addAction(node._action1)
        return ret

point_schema = StructSchema(MapPoint, MapPointRoles,
                            ("color", ColorSchema()),
                            ("r", IntSliderSchema("Radius")))

class MapProfileRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled | Qt.Qt.ItemFlag.ItemIsEditable | Qt.Qt.ItemFlag.ItemIsUserCheckable
    @staticmethod
    def field_roles(field):
        if field == "color":
            return [Qt.Qt.ItemDataRole.DecorationRole]
        if field == "name":
            return [Qt.Qt.ItemDataRole.DisplayRole]
        if field == "show_in_map":
            return [Qt.Qt.ItemDataRole.CheckStateRole]
        return None

    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return model.name
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return model.color
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            return Qt.Qt.CheckState.Checked if model.show_in_map else Qt.Qt.CheckState.Unchecked
    @staticmethod
    def set_role(model, role, value):
        if role == Qt.Qt.ItemDataRole.EditRole:
            model.name = value
            return True
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            model.show_in_map = (value == Qt.Qt.CheckState.Checked.value)
            return True
        return False

    @staticmethod
    def context_menu(parent_widget, node):
        ret = Qt.QMenu(parent_widget)
        node._action1 = Qt.QAction("Remove")
        def remove_self():
            me = node.self_index()
            if me.isValid():
                me.model().removeRow(me.row(), node.parent_index())
        node._action1.triggered.connect(remove_self)
        ret.addAction(node._action1)
        return ret

profile_schema = StructSchema(MapProfile, MapProfileRoles,
                              ("color", ColorSchema()),
                              ("r", IntSliderSchema("Radius")),
                              ("focus_point", FloatSliderSchema("Focus")))

class LayerRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled | Qt.Qt.ItemFlag.ItemIsUserCheckable
    @staticmethod
    def field_roles(field):
        if field == "is_enabled":
            return [Qt.Qt.ItemDataRole.CheckStateRole]
        return None
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            return Qt.Qt.Checked if model.is_enabled else Qt.Qt.Unchecked
        return None
    @staticmethod
    def set_role(model, role, value):
        if role == Qt.Qt.ItemDataRole.CheckStateRole:
            model.is_enabled = value == Qt.Qt.Checked.value
            return True
        return False

class PointsLayerRoles(LayerRoles):
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Points Layer"
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return Qt.QIcon('insarviz:points.png')
        return LayerRoles.get_role(model, role)

points_layer_schema = StructSchema(PointsLayer, PointsLayerRoles,
                                   ("opacity", FloatSliderSchema("Opacity")))

class SelectedBandLayerRoles(LayerRoles):
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Selected band layer"
        return LayerRoles.get_role(model, role)

selected_band_layer_schema = StructSchema(SelectedBandLayer, SelectedBandLayerRoles,
                                          ("opacity", FloatSliderSchema("Opacity")),
                                          ("dem_weight", FloatSliderSchema("DEM scale")))

class SwipeLayerRoles(LayerRoles):
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Swipe layer"
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return Qt.QIcon('insarviz:swipetool.png')
        return LayerRoles.get_role(model, role)
    @staticmethod
    def context_menu(parent_widget, node):
        ret = Qt.QMenu(parent_widget)
        node._action1 = Qt.QAction("Remove")
        def remove_self():
            me = node.self_index()
            if me.isValid():
                me.model().removeRow(me.row(), node.parent_index())
        node._action1.triggered.connect(remove_self)
        ret.addAction(node._action1)
        return ret
swipe_layer_schema = StructSchema(SwipeLayer, SwipeLayerRoles,
                                  ("cutoff", FloatSliderSchema("Cutoff")))

class RasterLayerRoles(LayerRoles):
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Raster layer"
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return Qt.QIcon('insarviz:raster1B.svg')
        return LayerRoles.get_role(model, role)

    @staticmethod
    def context_menu(parent_widget, node):
        ret = Qt.QMenu(parent_widget)
        node._action1 = Qt.QAction("Remove")
        def remove_self():
            me = node.self_index()
            if me.isValid():
                me.model().removeRow(me.row(), node.parent_index())
        node._action1.triggered.connect(remove_self)
        ret.addAction(node._action1)
        return ret
raster_layer_schema = StructSchema(RasterLayer, RasterLayerRoles,
                                   ("description", StringSchema(is_editable=False)),
                                   ("colormap", ColorMapSchema("colormap")),
                                   ("opacity", FloatSliderSchema("Opacity")),
                                   ("dem_weight", FloatSliderSchema("DEM scale")),
                                   )

class RasterRGBLayerRoles(LayerRoles):
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Raster RGB layer"
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return Qt.QIcon('insarviz:rasterRGB.svg')
        return LayerRoles.get_role(model, role)

    @staticmethod
    def context_menu(parent_widget, node):
        ret = Qt.QMenu(parent_widget)
        node._action1 = Qt.QAction("Remove")
        def remove_self():
            me = node.self_index()
            if me.isValid():
                me.model().removeRow(me.row(), node.parent_index())
        node._action1.triggered.connect(remove_self)
        ret.addAction(node._action1)
        return ret
raster_rgb_layer_schema = StructSchema(RasterRGBLayer, RasterRGBLayerRoles,
                                      ("description", StringSchema(is_editable=False)),
                                      ("opacity", FloatSliderSchema("Opacity")),
                                      )

class LayerListMenu(Qt.QMenu):
    def __init__(self, parent_widget, layer_list):
        super().__init__(parent_widget)
        self._layer_list = layer_list
        self._add_raster_layer_act = Qt.QAction(Qt.QIcon('insarviz:raster1B.svg'), "Add Raster Layer...")
        self._add_raster_layer_act.triggered.connect(self._add_raster_layer)
        self.addAction(self._add_raster_layer_act)
        self._add_raster_rgb_layer_act = Qt.QAction(Qt.QIcon('insarviz:rasterRGB.svg'), "Add Raster RGB Layer...")
        self._add_raster_rgb_layer_act.triggered.connect(self._add_raster_rgb_layer)
        self.addAction(self._add_raster_rgb_layer_act)
        self._add_swipe_layer_act = Qt.QAction(Qt.QIcon('insarviz:swipetool.png'), "Add Swipe Layer")
        self._add_swipe_layer_act.triggered.connect(self._add_swipe_layer)
        self.addAction(self._add_swipe_layer_act)

    @Qt.Slot()
    def _add_raster_layer(self):
        dataset_file, _ = Qt.QFileDialog.getOpenFileName(self.parent(), "Open Dataset", os.getcwd(), "Dataset (*.tiff *.tif *.h5 depl_cumule)")
        if dataset_file is None or dataset_file == "":
            logger.warn("Raster dataset not provided. Aborting raster layer creation")
            return
        dataset = Dataset(dataset_file)
        dialog = RasterLayerDialog(dataset)
        dialog.exec()
        layer = RasterLayer(dataset, dialog.band_number)
        self._layer_list.append(layer)

    @Qt.Slot()
    def _add_raster_rgb_layer(self):
        dataset_file, _ = Qt.QFileDialog.getOpenFileName(self.parent(), "Open RGB Dataset", os.getcwd(), "Dataset (*.tiff *.tif *.h5 depl_cumule)")
        if dataset_file is None or dataset_file == "":
            logger.warn("Raster RGB dataset not provided. Aborting RGB raster layer creation")
            return
        dataset = Dataset(dataset_file)
        layer = RasterRGBLayer(dataset)
        self._layer_list.append(layer)

    @Qt.Slot()
    def _add_swipe_layer(self):
        self._layer_list.append(SwipeLayer())

class LayerListRoles(ListRoles):
    def __init__(self):
        super().__init__("Layers")

    def context_menu(self, parent_widget, node):
        return LayerListMenu(parent_widget, node.model)

class TileProviderRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Tile provider: {model.name}"

class MaybeBandNumberRoles(StructRoles):
    @staticmethod
    def flags():
        return Qt.Qt.ItemFlag.ItemIsEnabled
    @staticmethod
    def get_role(model, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return "Reference band"

window_state_schema = StructSchema(WindowState, StructRoles,
                                   ("band_number", IntSchema("Current band")),
                                   ("maybe_reference_band", MaybeStructSchema(MaybeBandNumber, MaybeBandNumberRoles,
                                                                              ("band_number", IntSchema("band")))),
                                   ("points", ListSchema(ListIconRoles("Points", "insarviz:points.png"), point_schema)),
                                   ("profiles", ListSchema(ListIconRoles("Profiles", "insarviz:profile.png"), profile_schema)),
                                   ("layers", ListSchema(LayerListRoles(), UnionSchema(
                                       (PointsLayer, points_layer_schema),
                                       (SelectedBandLayer, selected_band_layer_schema),
                                       (RasterLayer, raster_layer_schema),
                                       (RasterRGBLayer, raster_rgb_layer_schema),
                                       (SwipeLayer, swipe_layer_schema),
                                   ))),
                                   ("tile_provider", TileProviderSchema()))

class LayerActions(Qt.QActionGroup):
    request_selection_change = Qt.Signal(Qt.QModelIndex)

    def __init__(self, parent, layers, points, profiles):
        super().__init__(parent)
        self.__layers = layers
        self.__points = points
        self.__profiles = profiles
        self.__selected_item = None
        self.__item_list = None

        self.__show_all_action = Qt.QAction(Qt.QIcon("insarviz:eye_open.svg"), "Show All", self)
        self.__show_all_action.triggered.connect(self.__show_all_layers)
        self.__hide_all_action = Qt.QAction(Qt.QIcon("insarviz:eye_closed.svg"), "Hide All", self)
        self.__hide_all_action.triggered.connect(self.__hide_all_layers)
        self.__move_up_action = Qt.QAction(Qt.QIcon("insarviz:arrowup.png"), "Move Up", self)
        self.__move_up_action.triggered.connect(lambda: self.__move_item(-1))
        self.__move_down_action = Qt.QAction(Qt.QIcon("insarviz:arrowdown.png"), "Move Down", self)
        self.__move_down_action.triggered.connect(lambda: self.__move_item(1))
        self.__delete_action = Qt.QAction(Qt.QIcon("insarviz:remove.png"), "Remove", self)
        self.__delete_action.triggered.connect(self.__remove_item)

        self.addAction(self.__show_all_action)
        self.addAction(self.__hide_all_action)
        self.addAction(self.__move_up_action)
        self.addAction(self.__move_down_action)
        self.addAction(self.__delete_action)
        self.set_selected_item(None)

    def __show_all_layers(self):
        for l in self.__layers:
            l.is_enabled = True
    def __hide_all_layers(self):
        for l in self.__layers:
            l.is_enabled = False
    def set_selected_item(self, item):
        enable = True
        self.__selected_item = item
        is_removable = True
        if item is not None:
            if isinstance(item.model, Layer):
                self.__item_list = self.__layers
                is_removable = item.model.is_removable
            elif isinstance(item.model, MapPoint):
                self.__item_list = self.__points
            elif isinstance(item.model, MapProfile):
                self.__item_list = self.__profiles
            else:
                enable = False
                self.__selected_item = None
                self.__item_list = None
        else:
            enable = False
        self.__move_up_action.setEnabled(enable)
        self.__move_down_action.setEnabled(enable)
        self.__delete_action.setEnabled(enable and is_removable)

    @Qt.Slot()
    def __remove_item(self):
        item = self.__selected_item
        if item is None:
            return
        index = item.self_index().row()
        self.__item_list[index:index+1] = []

    def __move_item(self, delta):
        item = self.__selected_item
        if item is None:
            return
        parent_list = self.__item_list
        index = item.self_index().row()

        if index+delta < 0 or index+delta >= len(parent_list):
            return
        item_model = item.item_model()
        parent = item.parent_index()

        elem = parent_list[index]
        parent_list[index:index+1] = []
        parent_list[index+delta:index+delta] = [elem]

        new_index = item_model.index(index+delta, 0, parent)
        self.set_selected_item(item_model.node_at(new_index))
        self.request_selection_change.emit(new_index)

class MapManagerWidget(Qt.QWidget):
    class ChildrenOfInterest:
        def __init__(self):
            self.state_widget: SchemaView
            self.toolbar: LayerActions

    def __init__(self, window_state):
        super().__init__()
        self.__layer_actions = LayerActions(self, window_state.layers, window_state.points, window_state.profiles)

        wtree = Container(Qt.QVBoxLayout, (), [
            Leaf(Qt.QToolBar, (), id="toolbar"),
            (Leaf(SchemaView, (window_state_schema, window_state), id="state_widget"), {"stretch":1})
        ])
        self.widgets = self.ChildrenOfInterest()
        wtree.create_in(self, self.widgets)
        self.widgets.state_widget.selection_changed.connect(self.__on_selection_changed)

        self.layer_actions.request_selection_change.connect(self.__request_selection_change)
        for act in self.layer_actions.actions():
            self.widgets.toolbar.addAction(act)

    @property
    def layer_actions(self):
        return self.__layer_actions

    @Qt.Slot()
    def __on_selection_changed(self, item):
        self.layer_actions.set_selected_item(item)
    @Qt.Slot()
    def __request_selection_change(self, index):
        self.widgets.state_widget.selectionModel().select(index, Qt.QItemSelectionModel.Select)



import numpy as np
from OpenGL import GL
import math

from .__prelude__ import (
    SelectedBand, Qt, Shader, Point, Matrix, MapPoint, MapProfile, Bound,
    inOpenGLContext, OpenGLPainter,
    dynamic, SELF,
    GLProgram, PureShader,
    logger
)
from .Scene import Scene, SceneDrawing
from .layer import LayerStack
from .TileCache import TileCache

MAP_SHADER = """
#version 330
float {geo_height}(vec3 geo_coords);
vec4 {geo_color}(vec3 geo_coords, inout float up_opacity);

float geo_height(vec3 geo_coords) {{
  return {geo_height}(geo_coords);
}}
vec4 geo_color(vec3 geo_coords) {{
  float up_opacity = 1.0;
  return {geo_color}(geo_coords, up_opacity);
}}
"""

class MapScene(Scene):
    selected_band     = dynamic.readonly()
    band_colormap     = dynamic.readonly()
    tile_cache        = dynamic.readonly()
    temp_point        = dynamic.readonly()
    temp_profile      = dynamic.readonly()
    overlay_layer     = dynamic.readonly()
    flip_horizontally = dynamic.variable(False)
    flip_vertically   = dynamic.variable(False)
    geo_focus         = dynamic.variable(None)

    def __init__(self, selected_band: SelectedBand, points, profiles, layers, band_colormap):
        super().__init__()
        self._selected_band = selected_band
        self._shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex,"map/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.TessellationControl,"map/tess_control.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, "map/tess_eval.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, "map/utils.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"map/fragment.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"hsv_rgb.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"map/utils.glsl")
        ]
        self._tile_shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex,"tile/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"tile/fragment.glsl"),
        ]
        self._band_colormap = band_colormap

        self._tile_cache = TileCache()
        self.tile_cache.new_tile_available.connect(self.renderChanged.emit)

        self.color = (0.5, 0.5, 0.5, 0.5)

        self.distance = 5.0
        self.__points = points
        self.__profiles = profiles

        self._overlay_layer = LayerStack(
            SELF.selected_band.dataset[self],
            SELF.temp_point[self],
            SELF.temp_profile[self],
            layers
        )
        self._overlay_layer.renderChanged.connect(self.renderChanged.emit)
        self._temp_point = None
        self._temp_profile = None

        self.dynamic_attribute("_rerender").value_changed.connect(lambda: self._rerender())

    def init_from_dict(self, dct):
        self.yaw = dct['yaw']
        self.pitch = dct['pitch']
        self.distance = dct['distance']
        self.center = Point(*dct['center'])
        self.heightUnits = dct['heightUnits']
        self.flip_horizontally = dct.get('flip_horizontally', False)
        self.flip_vertically = dct.get('flip_vertically', False)

    def to_dict(self):
        return {
            'yaw': self.yaw,
            'pitch': self.pitch,
            'distance': self.distance,
            'center': [self.center.x, self.center.y],
            'heightUnits': self.heightUnits,
            'flip_horizontally': self.flip_horizontally,
            'flip_vertically': self.flip_vertically,
        }

    @dynamic.memo(SELF.distance, SELF.pitch, SELF.yaw, SELF.center)
    def model_to_world(self):
        return Matrix.product(
            Matrix.translate((0.0,0.0,self.distance)),
            Matrix.rotate3(self.pitch, (-1.0, 0.0, 0.0, 0.0)),
            Matrix.rotate3(self.yaw, (0.0, 0.0, 1.0, 0.0)),
            Matrix.translate((self.center.x, self.center.y, 0.0))
        )
    @dynamic.memo(SELF.flip_horizontally, SELF.flip_vertically)
    def model_to_texture(self):
        h_flip = -1.0 if self.flip_horizontally else 1.0
        v_flip = -1.0 if self.flip_vertically else 1.0
        return Matrix.product(
            Matrix.scale((h_flip, v_flip, 1.0)),
            Matrix.scale((0.5,0.5,1.0)),
            Matrix.translate((h_flip,v_flip))
        )

    @dynamic.memo(SELF.band_colormap.xzero, SELF.band_colormap.xone)
    def value_to_elevation(self):
        lower, upper = self.band_colormap.xzero, self.band_colormap.xone
        scale = max(abs(lower), abs(upper))
        return Matrix.scale((1./scale, 1.0))
    @dynamic.memo(SELF.selected_band.dataset)
    def image_to_geo(self):
        w,h = self.selected_band.size
        return Matrix.product(
            self.selected_band.dataset.pixel_to_crs,
            Matrix.scale((w,h,1))
        )

    @dynamic.memo(SELF.selected_band.dataset)
    def show_background(self):
        return self.selected_band.dataset.is_georeferenced

    @dynamic.method(SELF.selected_band.image, SELF.selected_band.reference_image,
                    SELF.heightUnits, SELF.model_to_world, SELF.temp_point, SELF.temp_profile,
                    SELF.model_to_texture, SELF.geo_focus,
                    SELF.band_colormap.image, SELF.band_colormap.value_to_color)
    def _rerender(self):
        self.renderChanged.emit()

    def __create_point(self, dragStart: tuple[int, int], dragEnd: tuple[int, int]):
        xa,ya = dragStart
        xb,yb = dragEnd
        model_to_image = Matrix.product(
            self.selected_band.texture.texture_to_image,
            self.model_to_texture
        )
        def sq(x):
            return x*x
        dx_image, dy_image, _ = model_to_image.transform_vect((xb-xa, yb-ya, 0.0))
        w,h = self.selected_band.size
        # Clamp the radius to 100 pixels, to avoid users accidentally
        # computing the mean of the whole dataset
        radius = min(int(math.sqrt(sq(dx_image*w) + sq(dy_image*h))), 100)
        x_image, y_image = model_to_image.transform_point(dragStart)
        return MapPoint("Point", x_image*w, y_image*h, radius, Qt.QColor("blue"))
    def __create_profile(self, dragStart):
        dragEnd = dragStart
        x0, y0 = self.__model_to_pixel(*dragStart)
        x1, y1 = self.__model_to_pixel(*dragEnd)
        return MapProfile("Profile", 0, Qt.QColor("orange"), True, x0, y0, [(x1, y1)])
    def __model_to_pixel(self, x_model, y_model):
        model_to_image = Matrix.product(
            self.selected_band.texture.texture_to_image,
            self.model_to_texture
        )
        x_image, y_image = model_to_image.transform_point((x_model, y_model))
        w , h = self.selected_band.size
        return (int(x_image*w), int(y_image*h))

    def startDrag(self, dragMode, dragStart):
        if dragMode == Qt.Qt.KeyboardModifier.ShiftModifier:
            if self.temp_profile is None:
                self._temp_profile = self.__create_profile(dragStart)
            else:
                self.__append_profile_point(dragStart)
        return dragStart
    def whileDrag(self, dragMode, dragStart, dragEnd):
        if dragMode == Qt.Qt.KeyboardModifier.NoModifier:
            xa, ya = dragStart
            xb, yb = dragEnd
            self.center += Point(xb-xa, yb-ya)
        if dragMode == Qt.Qt.KeyboardModifier.ControlModifier:
            self._temp_point = self.__create_point(dragStart, dragEnd)
        if dragMode == Qt.Qt.KeyboardModifier.ShiftModifier:
            self._temp_profile.end_points[-1] = self.__model_to_pixel(*dragEnd)
            self.dynamic_attribute("temp_profile").value_changed.emit(self.temp_profile)

        return dragStart
    def endDrag(self, dragMode, dragStart, dragEnd):
        if dragMode == Qt.Qt.KeyboardModifier.ControlModifier:
            self._temp_point = None
            self.__points.append(self.__create_point(dragStart, dragEnd))
        if dragMode == Qt.Qt.KeyboardModifier.ShiftModifier:
            if self._temp_profile is not None:
                self.__append_profile_point(dragEnd)
            return dragEnd
    def __append_profile_point(self, pt):
        new_point = self.__model_to_pixel(*pt)
        if len(self.temp_profile.end_points) < 2 or new_point != self.temp_profile.end_points[-2]:
            self.temp_profile.end_points.append(new_point)
        self.dynamic_attribute("temp_profile").value_changed.emit(self.temp_profile)

    def abortDrag(self, dragMode, dragStart):
        if dragMode == Qt.Qt.KeyboardModifier.ControlModifier:
            self._temp_point = None
        if dragMode == Qt.Qt.KeyboardModifier.ShiftModifier:
            if len(self.temp_profile.end_points) > 1:
                self.temp_profile.end_points[-1:] = []
                self.__profiles.append(self.temp_profile)
                self._temp_profile = None

    def initializeDrawing(self, context):
        gl = context.functions()
        logger.debug("Loading map scene (max texture size : %d)", gl.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE))
        return InitializedMapScene(context, self)

class InitializedMapScene(SceneDrawing["MapScene"]):
    def __init__(self, context, map_scene):
        super().__init__(context, map_scene)

        self.__tile_program: Qt.QOpenGLShaderProgram = GLProgram(map_scene._tile_shaders)
        self.__tile_mesh = self.__tile_program.create_square_mesh(GL.GL_TRIANGLES, split=0, texture_to_model = Matrix.identity(3))

        self.__initialized_overlay = self.drawing.overlay_layer.GL_initialize(context)
        GL.glPatchParameteri(GL.GL_PATCH_VERTICES, 3)

    @dynamic.memo(SELF.drawing.overlay_layer.shaders)
    def GL_program(self):
        return GLProgram(self.drawing._shaders + [
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, shader)
            for shader in self.drawing.overlay_layer.shaders
        ] + [
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, shader)
            for shader in self.drawing.overlay_layer.shaders
        ] + [
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.Fragment, MAP_SHADER.format(
                geo_color = self.drawing.overlay_layer.local_ident("geo_color"),
                geo_height = self.drawing.overlay_layer.local_ident("geo_height")
            )),
            PureShader(Qt.QOpenGLShader.ShaderTypeBit.TessellationEvaluation, MAP_SHADER.format(
                geo_color = self.drawing.overlay_layer.local_ident("geo_color"),
                geo_height = self.drawing.overlay_layer.local_ident("geo_height")
            ))

        ])
    @dynamic.memo(SELF.GL_program, SELF.drawing.model_to_texture)
    def GL_mesh(self):
        return self.GL_program.create_square_mesh(
            GL.GL_PATCHES, split=8,
            texture_to_model = self.drawing.model_to_texture.inverse())

    @dynamic.memo(SELF.viewport_size)
    def world_to_clip(self):
        w,h = self.viewport_size
        hfactor = max(1.0, w/h)/4.0
        vfactor = max(1.0, h/w)/4.0
        near = 4.0*0.005
        return Matrix.frustum(-hfactor*near, hfactor*near,
                              -vfactor*near, vfactor*near,
                              -near, -1000.0)
    @dynamic.memo(SELF.viewport_size)
    def tiles_per_screen(self):
        w,h = self.viewport_size
        return max(w/256.0, h/256.0) # Each tile is 256 pixels wide
    @dynamic.memo(SELF.viewport_size)
    def clip_to_pixel(self):
        w,h = self.viewport_size
        return Matrix.scale((w/2.0, -h/2.0, 1.0)) * Matrix.translate((1.0,-1.0))

    def paintGL(self):
        program = self.GL_program
        mesh = self.GL_mesh
        glfunc = self.context.functions()
        setup_overlay = self.__initialized_overlay.GL_setup

        glfunc.glEnable(GL.GL_DEPTH_TEST)

        glfunc.glClearColor(*self.drawing.color)
        glfunc.glClear(int(GL.GL_COLOR_BUFFER_BIT) | int(GL.GL_DEPTH_BUFFER_BIT))

        if self.drawing.show_background:
            if self.drawing.pitch < 60.0:
                self.drawing.tile_cache.GL_paint_background(
                    glfunc,
                    model_to_crs = Matrix.product(
                        self.drawing.selected_band.image_to_crs,
                        self.drawing.selected_band.texture.texture_to_image,
                        self.drawing.model_to_texture
                    ),
                    clip_to_model = self.clip_to_model,
                    model_to_world = self.drawing.model_to_world,
                    world_to_clip = self.world_to_clip,
                    tiles_per_screen = self.tiles_per_screen,
                )

        dem_texture = self.drawing.selected_band.texture
        ref_texture = self.drawing.selected_band.reference_texture
        tex_to_image = dem_texture.texture_to_image

        glfunc.glEnable(GL.GL_BLEND)
        glfunc.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        with program as p:
            p.uniforms.image = dem_texture
            p.uniforms.ref_image = ref_texture

            p.uniforms.texture_to_image = tex_to_image
            p.uniforms.image_to_texture = tex_to_image.inverse()

            p.uniforms.image_to_geo = self.drawing.image_to_geo
            p.uniforms.geo_to_model = Matrix.product(
                self.drawing.image_to_geo,
                tex_to_image,
                self.drawing.model_to_texture,
            ).inverse()
            setup_overlay(p)

            p.uniforms.heightUnits = float(self.drawing.heightUnits)
            p.uniforms.value_to_elevation = self.drawing.value_to_elevation
            _,_,z_near = self.world_to_clip.inverse().transform_point((0.0,0.0,-1.0))
            p.uniforms.near = z_near

            p.uniforms.world_to_clip = self.world_to_clip
            p.uniforms.model_to_world = self.drawing.model_to_world

            mesh.draw()

        if self.drawing.geo_focus is not None:
            with OpenGLPainter(*self.viewport_size) as painter:
                center = (tex_to_image * self.drawing.model_to_texture).inverse().transform_point(self.drawing.geo_focus)
                clip_cx, clip_cy, _ = (self.world_to_clip * self.drawing.model_to_world).transform_point((*center, 0.0))
                pixel_cx, pixel_cy = self.clip_to_pixel.transform_point((clip_cx, -clip_cy))
                painter.setBrush(Qt.QColor('red'))
                painter.drawEllipse(Qt.QPointF(pixel_cx, pixel_cy), 10 , 10)

    def freeGL(self):
        self.__initialized_overlay.GL_free()



from .__prelude__ import Qt
from .Schema import Schema

class FloatSliderSchema(Schema):
    class FloatSliderWidget(Qt.QGroupBox):
        valueChanged = Qt.Signal(float)

        def __init__(self, name, f_min, f_max):
            super().__init__(name)

            layout = Qt.QHBoxLayout(self)
            self._layout = layout
            self._f_min = f_min
            self._f_max = f_max

            slider = Qt.QSlider(Qt.Qt.Orientation.Horizontal)
            slider.setMinimum(0)
            slider.setMaximum(1000)
            slider.valueChanged.connect(self._on_value_changed)
            layout.addWidget(slider, stretch=1)
            self._slider : Qt.QSlider = slider
            layout.setContentsMargins(5,2,5,2)
            self.setLayout(layout)

        @Qt.Slot(int)
        def _on_value_changed(self, __n__):
            self.valueChanged.emit(self.value())

        def value(self):
            return float(self._slider.value()) * 0.001 * (self._f_max - self._f_min) + self._f_min
        def setValue(self, value: float):
            return self._slider.setValue(int((value - self._f_min) * 1000.0 / (self._f_max - self._f_min)))

    class FloatSliderNode(Schema.Node):
        def __init__(self, item_model, schema, model, name):
            super().__init__(item_model, schema)
            self._create_editor()
            self._init_model(model)

        def _create_editor(self):
            self._editor = FloatSliderSchema.FloatSliderWidget(self._schema.name, self._schema.f_min, self._schema.f_max)
            self._editor.valueChanged.connect(self.requestModelChange.emit)

        def createEditor(self, parent):
            self._editor.setParent(parent)
            return self._editor
        def destroyEditor(self):
            return False
        def editor(self):
            return self._editor
        def setModelData(self, __editor__):
            return True
        def setEditorData(self, editor):
            editor.setValue(self.model)
            return True

        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsEditable | super().flags()
        def _init_model(self, model):
            super()._init_model(model)
            self.dataChanged.emit([])

        # def data(self, role):
        #     if role == Qt.ItemDataRole.DisplayRole:
        #         return self.model
        #     return None

    def __init__(self, name, f_min = 0.0, f_max = 1.0):
        self.name = name
        self.f_min = f_min
        self.f_max = f_max

    def make_node(self, item_model, model):
        return self.FloatSliderNode(item_model, self, model, self.name)

class FloatSchema(Schema):
    class FloatNode(Schema.Node):
        def __init__(self, item_model, schema, model, name):
            super().__init__(item_model, schema)
            self.name = name
            self._create_editor()
            self._init_model(model)

        def _create_editor(self):
            self._editor = Qt.QDoubleSpinBox()
            self._editor.setPrefix(f"{self.name} : ")
            self._editor.valueChanged.connect(self.requestModelChange.emit)

        def createEditor(self, parent):
            self._editor.setParent(parent)
            return self._editor
        def destroyEditor(self):
            return False
        def editor(self):
            return self._editor
        def setModelData(self, __editor__):
            return True
        def setEditorData(self, editor):
            editor.setValue(self.model)
            return True

        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsEditable | super().flags()
        def _init_model(self, model):
            super()._init_model(model)
            self.dataChanged.emit([])

    def __init__(self, name):
        self.name = name

    def make_node(self, item_model, model):
        return self.FloatNode(item_model, self, model, self.name)



from typing import Any
import numpy as np
import pyqtgraph
import pyqtgraph.opengl
from pyqtgraph.GraphicsScene import exportDialog

from .__prelude__ import (
    Qt, color_icon, PointData, ProfileData, linmap, ProfileScene,
    Container, Leaf,
    ItemChooser,
    ComputedProgressBar,
    IconLabel,
    GLWidget,
    SELF,
    BoldLabel, icon_with_label
)

class ProfileDataItem2D(pyqtgraph.PlotItem):
    def __init__(self, profile_data):
        super().__init__()
        self.__profile_data = profile_data
        self.__scatter_item = pyqtgraph.PlotDataItem(symbolSize=5)
        # self.scatter_item.scatter.setData(hoverable=True)
        self.addItem(self.scatter_item)

        self.__spatial_std_curve_low = pyqtgraph.PlotCurveItem(connect='finite', pen=self.__profile_data.color)
        self.__spatial_std_curve_high = pyqtgraph.PlotCurveItem(connect='finite', pen=self.__profile_data.color)
        self.__spatial_error_item = pyqtgraph.FillBetweenItem(self.__spatial_std_curve_low, self.__spatial_std_curve_high)
        self.spatial_error_item.setZValue(-1.0)
        self.addItem(self.spatial_error_item)

        self.__temporal_std_curve_low = pyqtgraph.PlotCurveItem(connect='finite', pen=self.__profile_data.color)
        self.__temporal_std_curve_high = pyqtgraph.PlotCurveItem(connect='finite', pen=self.__profile_data.color)
        self.__temporal_error_item = pyqtgraph.FillBetweenItem(self.__temporal_std_curve_low, self.__temporal_std_curve_high)
        self.temporal_error_item.setZValue(-1.0)
        self.addItem(self.temporal_error_item)

        self.__focus_line = pyqtgraph.InfiniteLine(angle=90, movable=True, pen=self.__profile_data.color)
        self.__focus_line.sigDragged.connect(self.__on_focus_dragged)
        self.addItem(self.__focus_line)

        self.__profile_data.dynamic_attribute("values_and_variance_2d").drive(self.__update_plot)
        self.__profile_data.dynamic_attribute("values_2d_temporal").drive(self.__update_temporal_plot)
        self.__profile_data.dynamic_attribute("color").drive(self.__set_color)
        self.__profile_data.dynamic_attribute("show_spatial_variance").drive(self.__set_show_spatial_variance)
        self.__profile_data.dynamic_attribute("show_temporal_variance").drive(self.__set_show_temporal_variance)
        self.__profile_data.dynamic_attribute("focus_point").drive(self.__set_focus_line_pos)

    @property
    def computed_values(self):
        return self.__profile_data.computed_values

    @property
    def scatter_item(self):
        return self.__scatter_item
    @property
    def spatial_error_item(self):
        return self.__spatial_error_item
    @property
    def temporal_error_item(self):
        return self.__temporal_error_item

    @Qt.Slot()
    def __on_focus_dragged(self):
        values = self.__profile_data.values_2d
        if values is None:
            return
        xs, _ = values
        i = int(self.__focus_line.getPos()[0])
        self.__profile_data.focus_point = max(0.0, min(1.0, i / (len(xs)-1)))

    @Qt.Slot()
    def __set_color(self, color):
        self.scatter_item.setPen(color)
        self.scatter_item.setSymbolPen(color)
        self.scatter_item.setBrush(color)
        self.scatter_item.setSymbolBrush(color)
        muted_color = Qt.QColor(color)
        muted_color.setAlpha(96)
        self.spatial_error_item.setBrush(muted_color)
        self.temporal_error_item.setBrush(muted_color)
        self.__focus_line.setPen(color)

    @Qt.Slot()
    def __set_show_spatial_variance(self, show_variance):
        self.spatial_error_item.setVisible(show_variance)
    @Qt.Slot()
    def __set_show_temporal_variance(self, show_variance):
        self.temporal_error_item.setVisible(show_variance)

    @Qt.Slot()
    def __set_focus_line_pos(self, focus_point):
        values = self.__profile_data.values_2d
        if values is None:
            return
        xs, _ = values
        index = int(focus_point * (len(xs)-1))
        self.__focus_line.setPos(index)

    @Qt.Slot()
    def __update_temporal_plot(self, values):
        if values is not None:
            xs, ys, temp_var = values
            temp_stddev = np.sqrt(temp_var)
            self.__temporal_std_curve_low.setData(xs, ys - temp_stddev)
            self.__temporal_std_curve_high.setData(xs, ys + temp_stddev)

    @Qt.Slot()
    def __update_plot(self, values):
        if values is not None:
            xs, ys, spatial_var = values
            self.scatter_item.setData(x = xs, y = ys)
            spatial_std_dev = np.sqrt(spatial_var)
            self.__spatial_std_curve_low.setData(xs, ys - spatial_std_dev)
            self.__spatial_std_curve_high.setData(xs, ys + spatial_std_dev)

class ProfilePlotWidget2D(pyqtgraph.PlotWidget):
    def __init__(self):
        super().__init__()
        self.__item = None
    def set_item(self, item):
        self.clear()
        self.__item = item
        if item is not None:
            for sub_item in item.items:
                self.addItem(sub_item)

def ProfileDataItem3D(profile_data):
    return profile_data
class ProfilePlotWidget3D(GLWidget):
    def __init__(self):
        super().__init__(ProfileScene(None))

    def set_item(self, item):
        self.scene.profile_data = item

class SpatialPlotWidget(Qt.QWidget):
    class ChildrenOfInterest:
        def __init__(self):
            self.plot_widget: pyqtgraph.PlotWidget
            self.profile_chooser: ItemChooser
            self.lock_axes_checkbox: Qt.QCheckBox
            self.export_button: Qt.QPushButton
            self.show_spatial_stddev_checkbox: Qt.QCheckBox
            self.show_temporal_stddev_checkbox: Qt.QCheckBox
            self.compute_progress: ComputedProgressBar
            self.reference_chooser: ItemChooser
            self.view_mode: Qt.QCheckBox
            self.plot_container: Qt.QWidget

    def __init__(self, selected_band, points, profiles):
        super().__init__()
        self.__selected_band = selected_band
        self.__points = points
        self.__profiles = profiles
        self.__profiles.endInsertRange.connect(self.__on_insert_profiles)
        self.__profiles.endRemoveRange.connect(self.__on_remove_profiles)
        self.__profiles.endReplaceRange.connect(self.__on_replace_profiles)
        self.__reference_data = PointData(self.__selected_band, None)
        self.__lock_axes = False
        self.__profile_data = [self.__make_profile_data(p) for p in self.__profiles]
        self._current_profile_data = None
        self.ProfileDataItem = ProfileDataItem2D
        self.widgets = self.ChildrenOfInterest()

        skel = Container(Qt.QVBoxLayout, (), [
            Container(Qt.QHBoxLayout, (), [
                Container(Qt.QStackedLayout, (), [
                    Leaf(ProfilePlotWidget2D, (), id="plot_widget_2d"),
                    Leaf(ProfilePlotWidget3D, (), id="plot_widget_3d")
                ], id="plot_container"),
                Container(Qt.QVBoxLayout, (), [
                    Leaf(Qt.QCheckBox, ("Lock Axes", ), id="lock_axes_checkbox"),
                    icon_with_label("insarviz:profile.png", "Profile"),
                    (Leaf(ItemChooser, (self.__profiles,), id="profile_chooser"), {"stretch": 1}),
                    Leaf(Qt.QCheckBox, ("Show spatial deviation", ), id="show_spatial_stddev_checkbox"),
                    Leaf(Qt.QCheckBox, ("Show temporal deviation", ), id="show_temporal_stddev_checkbox"),
                    icon_with_label("insarviz:points.png", "Reference"),
                    Leaf(ItemChooser, (self.__points,), id="reference_chooser"),
                    icon_with_label("insarviz:logo_insarviz.png", "View Mode"),
                    Leaf(Qt.QCheckBox, ("3D View (experimental)",), id="view_mode"),
                    Leaf(Qt.QSpinBox, (), id="smoothing_factor"),
                    (Leaf(Qt.QWidget, ()), {"stretch": 1}),
                    Leaf(Qt.QPushButton, ("Export...",), id="export_button"),
                ]),

            ]),
            Leaf(ComputedProgressBar, (), id="compute_progress")
        ])
        skel.create_in(self, self.widgets)

        self.widgets.smoothing_factor.valueChanged.connect(self.__on_smoothing_changed)
        self.widgets.smoothing_factor.setEnabled(False)
        self.widgets.view_mode.setChecked(False)
        self.widgets.view_mode.checkStateChanged.connect(self.__set_view_mode)
        self.widgets.profile_chooser.current_item_changed.connect(self.__on_profile_selected)
        self.widgets.lock_axes_checkbox.checkStateChanged.connect(self.__on_lock_axes)
        self.widgets.export_button.clicked.connect(self.__export)
        self.widgets.show_spatial_stddev_checkbox.checkStateChanged.connect(self.__on_show_spatial_error)
        self.widgets.show_temporal_stddev_checkbox.checkStateChanged.connect(self.__on_show_temporal_error)
        self.widgets.reference_chooser.current_item_changed.connect(self.__on_reference_selected)
        self.widgets.plot_widget = self.widgets.plot_widget_2d
        self.__dynamic_units = SELF.dataset.value_units[self.__selected_band]
        self.__dynamic_units.drive(lambda units: self.widgets.plot_widget_2d.setLabel('left', f"LOS displacement ({units})" if units is not None else "No units"))
        self.widgets.plot_widget_2d.setLabel('bottom', "Distance along profile (px)")
        self.widgets.plot_widget_2d.setBackground('w')
        self.widgets.plot_widget_2d.setForegroundBrush(Qt.Qt.GlobalColor.transparent)

        self.__auto_range()

    def __auto_range(self):
        self.widgets.plot_widget.autoRange()
        if self.__lock_axes:
            self.widgets.plot_widget.disableAutoRange()
        else:
            self.widgets.plot_widget.enableAutoRange()

    @Qt.Slot()
    def __export(self):
        self.__export_dialog = exportDialog.ExportDialog(self.widgets.plot_widget_2d.plotItem.scene())
        self.__export_dialog.show(self.widgets.plot_widget_2d)

    def __make_profile_data(self, profile):
        return ProfileData(self.__selected_band, profile, self.__reference_data)

    @Qt.Slot(Any)
    def __on_show_spatial_error(self, checked):
        show_bars = checked == Qt.Qt.Checked
        if show_bars:
            self.widgets.show_temporal_stddev_checkbox.setChecked(False)
        p_data = self._current_profile_data
        if p_data is not None:
            p_data.show_spatial_variance = show_bars
            self.__auto_range()
    @Qt.Slot(Any)
    def __on_show_temporal_error(self, checked):
        show_bars = checked == Qt.Qt.Checked
        if show_bars:
            self.widgets.show_spatial_stddev_checkbox.setChecked(False)
        p_data = self._current_profile_data
        if p_data is not None:
            p_data.show_temporal_variance = show_bars
            self.__auto_range()
    @Qt.Slot()
    def __on_smoothing_changed(self):
        if self._current_profile_data is not None:
            self._current_profile_data.smoothing_factor = self.widgets.smoothing_factor.value()
    @Qt.Slot()
    def __set_view_mode(self):
        if self.widgets.view_mode.isChecked():
            self.ProfileDataItem = ProfileDataItem3D
            self.widgets.plot_widget = self.widgets.plot_widget_3d
        else:
            self.ProfileDataItem = ProfileDataItem2D
            self.widgets.plot_widget = self.widgets.plot_widget_2d

        self.widgets.plot_container.layout().setCurrentWidget(self.widgets.plot_widget)
        self.__update_profile_widget()

    @Qt.Slot(Any)
    def __on_lock_axes(self, checked):
        self.__lock_axes = checked == Qt.Qt.Checked
        self.__auto_range()

    @Qt.Slot(int, int)
    def __on_insert_profiles(self, start, length):
        self.__profile_data[start:start] = [
            self.__make_profile_data(p)
            for p in self.__profiles[start:start+length]
        ]
    @Qt.Slot(int,int)
    def __on_remove_profiles(self, start, end):
        self.__profile_data[start:end] = []
    @Qt.Slot(int,int)
    def __on_replace_profiles(self, start, end):
        self.__profile_data[start:end] = [
            self.__make_profile_data(p)
            for p in self.__profiles[start:end]
        ]

    @Qt.Slot(int, Any)
    def __on_reference_selected(self, _, ref_point):
        self.__reference_data.set_point(ref_point)

    def __update_profile_widget(self):
        if self._current_profile_data is not None:
            profile_item = self.ProfileDataItem(self._current_profile_data)
            self.widgets.plot_widget.set_item(profile_item)
            self.widgets.compute_progress.set_computed_value(profile_item.computed_values)
        else:
            self.widgets.plot_widget.set_item(None)

    @Qt.Slot(int, Any)
    def __on_profile_selected(self, profile_index, profile):
        if profile is None:
            self._current_profile_data = None
            self.widgets.compute_progress.set_computed_value(None)
            self.widgets.smoothing_factor.setEnabled(False)
        else:
            self._current_profile_data = self.__profile_data[profile_index]
            self.widgets.smoothing_factor.setEnabled(True)
            self.widgets.smoothing_factor.setValue(self._current_profile_data.smoothing_factor)
            self.widgets.show_spatial_stddev_checkbox.setChecked(profile.show_spatial_variance)
            self.widgets.show_temporal_stddev_checkbox.setChecked(profile.show_temporal_variance)
        self.__update_profile_widget()



from typing import Optional
import math

from .__prelude__ import Qt, Matrix, Point, Scene, logger

class GLWidget(Qt.QOpenGLWidget):
    positionChanged = Qt.Signal(float, float)
    resized_viewport = Qt.Signal(int, int)

    def __init__(self, scene_info: Scene):
        super().__init__()
        self.setFocusPolicy(Qt.Qt.FocusPolicy.WheelFocus)
        self.setMouseTracking(True)
        self._scene: Scene  = scene_info
        self._scene.renderChanged.connect(self.update)
        self._initialized_scene = None
        self._initialized_scene_mods = []

        self._clip_to_pixel: Optional[Matrix]  = None
        # contains the position of the cursor (in model space) when dragging, and None otherwise
        self.dragMode = None
        self.dragStart: Optional[tuple[float, float]] = None

    @property
    def scene(self):
        return self._scene

    @property
    def clip_to_pixel(self) -> Matrix:
        if self._clip_to_pixel is None:
            w, h = float(self.width()), float(self.height())
            self._clip_to_pixel = Matrix.scale((w/2.0, -h/2.0, 1.0)) * Matrix.translate((1.0,-1.0))
        return self._clip_to_pixel

    @property
    def world_to_clip(self):
        return self.initialized_scene.world_to_clip

    def sizeHint(self):
        return Qt.QSize(100,100)

    def pixel_to_model(self, x, y) -> tuple[float,float]:
        clip_x, clip_y = self.clip_to_pixel.inverse().transform_point((x, y))
        return self.initialized_scene.clip_to_model(clip_x, clip_y)

    def mousePressEvent(self, event):
        pos = event.position()
        mods = Qt.QGuiApplication.keyboardModifiers()
        self.dragMode = mods
        self.dragStart = self._scene.startDrag(mods, self.pixel_to_model(pos.x(), pos.y()))
    def mouseReleaseEvent(self, event):
        if self.dragStart is not None:
            pos = event.position()
            self.dragStart = self._scene.endDrag(self.dragMode, self.dragStart, self.pixel_to_model(pos.x(), pos.y()))
    def mouseMoveEvent(self, event):
        pos = event.position()
        xb,yb = self.pixel_to_model(pos.x(), pos.y())
        if self.dragStart is not None:
            self.dragStart = self._scene.whileDrag(self.dragMode, self.dragStart, (xb,yb))
        self.positionChanged.emit(xb,yb)

    def keyPressEvent(self, event):
        if event.modifiers() == Qt.Qt.KeyboardModifier.NoModifier:
            if event.key() == Qt.Qt.Key.Key_Right:
                self._scene.selected_band.band_number += 1
            if event.key() == Qt.Qt.Key.Key_Left:
                self._scene.selected_band.band_number -= 1
    def keyReleaseEvent(self, event):
        if self.dragStart is not None:
            if event.modifiers() != self.dragMode:
                self._scene.abortDrag(self.dragMode, self.dragStart)
                self.dragStart = None

    def preserving_pointer(self, pos, doit):
        x_b, y_b = self.pixel_to_model(pos.x(), pos.y())
        doit()
        x_a, y_a = self.pixel_to_model(pos.x(), pos.y())
        self._scene.center += Point(x_a - x_b, y_a - y_b)

    def wheelEvent(self, event):
        angle = event.angleDelta().y() / 32.0
        mods = Qt.QGuiApplication.keyboardModifiers()
        if mods == Qt.Qt.KeyboardModifier.ShiftModifier | Qt.Qt.KeyboardModifier.ControlModifier:
            units = self._scene.heightUnits
            self._scene.heightUnits = min(1.0, max(0, units + angle/360.0))
        elif mods == Qt.Qt.KeyboardModifier.ShiftModifier:
            def doit():
                self._scene.pitch = max(0.0,min(85.0,self._scene.pitch + angle))
            self.preserving_pointer(event.position(), doit)
        elif mods == Qt.Qt.KeyboardModifier.ControlModifier:
            def doit():
                self._scene.yaw += angle
            self.preserving_pointer(event.position(), doit)
        else:
            def doit():
                self._scene.distance *= math.exp(math.log(0.25)*math.radians(angle))
            self.preserving_pointer(event.position(), doit)

    def modify_initialized_scene(self, mod):
        if self._initialized_scene is None:
            self._initialized_scene_mods.append(mod)
        else:
            mod(self._initialized_scene)

    @property
    def initialized_scene(self):
        if self._initialized_scene is None:
            self._initialized_scene = self._scene.initializeDrawing(self.context())
            for mod in self._initialized_scene_mods:
                mod(self._initialized_scene)
            self._initialized_scene_mods = []
        return self._initialized_scene

    def initializeGL(self):
        _ = self.initialized_scene

    def resizeGL(self, w, h):
        self._clip_to_pixel = None
        self.initialized_scene.resizeViewport(w, h)
        self.resized_viewport.emit(w,h)

    def paintGL(self):
        ret = self.initialized_scene.paintGL()
        return ret

    def closeEvent(self, event):
        event.accept()
        if self._initialized_scene is not None:
            self.context().makeCurrent(Qt.QOffscreenSurface())
            self._initialized_scene.freeGL()



from typing import TypeVar

from .__prelude__ import ObservableStruct, dynamic, Qt

GenericDrawing = TypeVar("GenericDrawing", covariant=True)

class InitializedDrawing[GenericDrawing](ObservableStruct):
    context = dynamic.readonly()
    drawing = dynamic.readonly()
    viewport_size = dynamic.readonly()

    def __init__(self, context: Qt.QOpenGLContext, drawing: GenericDrawing):
        super().__init__()
        self._context = context
        self._drawing = drawing
        self._viewport_size = (1,1)

    def resizeViewport(self, w, h, /):
        self._viewport_size = (w,h)
    def paintGL(self):
        pass
    def freeGL(self):
        pass

class GLDrawing(ObservableStruct):
    renderChanged = Qt.Signal()

    def initializeDrawing(self, context: Qt.QOpenGLContext, /) -> InitializedDrawing["GLDrawing"]:
        return InitializedDrawing(context, self)



import numpy as np

from .CurveParam import CurveParam
from .Curve import Curve

class ScaledStepCurve(Curve):
    def __init__(self):
        super().__init__()
        self.scale = CurveParam("scale")
        self.step = 0.0

    __mime_type__ = "application/x-insarviz/ScaledStepCurve"

    @property
    def nparams(self):
        return 1
    @property
    def params(self):
        return [self.scale]

    def func(self, xs, scale):
        return scale * np.where(xs >= self.step, xs, 0.0)
    def jacobian(self, xs, scale):
        return np.array([np.where(xs >= self.step, xs, 0.0)])
    def hessian_epsilon(self, xs, err, grad, v, b):
        return 0.0
